/* xtmix  - an X11 interface to all mixers supported by
 *          Hannu Savolainen's VoxWare (which is included 
 *          in most Linux distributions).
 *
 * (c) 1995 - 1998 by Martin Denn <Martin.Denn@rhein-main.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/* X stuff */
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/xpm.h>

/* Linux stuff */
#include <sys/types.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <stdio.h>
#include <math.h>
#include <linux/soundcard.h>
#include <sys/param.h>
#include <stdlib.h>

/* selfmade Widgets */
#include "Dial.h"
#include "Slide.h"

/* KDE support */
#include "kdesupport.h"

/* xpm-bitmaps */
#include "bitmaps/xtmix-mini.xpm"
#include "bitmaps/cd.xpm"
#include "bitmaps/synth.xpm"
#include "bitmaps/mic.xpm"
#include "bitmaps/pcm.xpm"
#include "bitmaps/line.xpm"
#include "bitmaps/speaker.xpm"
#include "bitmaps/empty.xpm"
#include "bitmaps/mix.xpm"
#include "bitmaps/pcm2.xpm"
#include "bitmaps/rec.xpm"
#include "bitmaps/igain.xpm"
#include "bitmaps/ogain.xpm"
#include "bitmaps/line1.xpm"
#include "bitmaps/line2.xpm"
#include "bitmaps/line3.xpm"

/* X bitmap to build the icon */
#include "bitmaps/xtmix.xbm"

/* misc defines */
#define VERSION		"0.4"
#define MIXERDEVICE     "/dev/mixer"
#define BALANCE		100 /* should not interfere with mixer device no */
#define MINI_ICON       101 /* should not interfere with mixer device no */
#define SQR(x)		(x*x)
#define CIRCLE(x)	(sqrt(SQR((50.0))-SQR((x-50.0))))
#define CORRECTION	5
#ifndef MAX
#define MAX(x,y) (x > y ? x : y)
#endif

/* Resource definitions local to xtmix */
#define XtNignoreKDE "ignoreKDE"
#define XtCIgnoreKDE "IgnoreKDE"
#define XtNlauncher "launcher"
#define XtCLauncher "Launcher"
#define XtNnoFilters "noFilters"
#define XtCNoFilters "NoFilters"
#define XtNmixerDevice "mixerDevice"
#define XtCMixerDevice "MixerDevice"
#define XtNlauncherBackground "launcherBackground"
#define XtCLauncherBackground "LauncherBackground"
#define XtNlauncherLabel "launcherLabel"
#define XtCLauncherLabel "LauncherLabel"
#define XtNrestore "restore"
#define XtCRestore "Restore"

typedef struct {
    Boolean launcher;
    Boolean ignore_KDE;
    Boolean no_filters;
    String mixer_device;
    String launcher_background;
    String launcher_label;
    String restore;
} AppData;

AppData app_data;

static XtResource resources[] = {
    {
        XtNlauncher,
        XtCLauncher,
        XtRBoolean,
        sizeof(Boolean),
        XtOffsetOf(AppData, launcher),
        XtRImmediate,
        (XtPointer) False,
    },
    {
        XtNignoreKDE,
        XtCIgnoreKDE,
        XtRBoolean,
        sizeof(Boolean),
        XtOffsetOf(AppData, ignore_KDE),
        XtRImmediate,
        (XtPointer) False,
    },
    {
        XtNnoFilters,
        XtCNoFilters,
        XtRBoolean,
        sizeof(Boolean),
        XtOffsetOf(AppData, no_filters),
        XtRImmediate,
        (XtPointer) False,
    },
    {
        XtNmixerDevice,
        XtCMixerDevice,
        XtRString,
        sizeof(String),
        XtOffsetOf(AppData, mixer_device),
        XtRString,
        MIXERDEVICE,
    },
    {
        XtNlauncherBackground,
        XtCLauncherBackground,
        XtRString,
        sizeof(String),
        XtOffsetOf(AppData, launcher_background),
        XtRString,
        NULL,
    },
    {
        XtNlauncherLabel,
        XtCLauncherLabel,
        XtRString,
        sizeof(String),
        XtOffsetOf(AppData, launcher_label),
        XtRString,
        NULL,
    },
    {
        XtNrestore,
        XtCRestore,
        XtRString,
        sizeof(String),
        XtOffsetOf(AppData, restore),
        XtRString,
        NULL,
    },
};

/* switch -> resource mapping */

static XrmOptionDescRec options[] = {
    {"-nofilters", "*noFilters", XrmoptionNoArg, "True"},
    {"-dev", "*mixerDevice", XrmoptionSepArg, NULL},
    {"-launch", "*launcher", XrmoptionNoArg, "True"},
    {"-lbg", "*launcherBackground", XrmoptionSepArg, NULL},
    {"-llabel", "*launcherLabel", XrmoptionSepArg, NULL},
    {"-restore", "*restore", XrmoptionSepArg, NULL},
    {"-ignoreKDE", "*ignoreKDE", XrmoptionNoArg, "True"},
};
        
/* Action Prototypes */
void messages(), quitApplication(), enterWindow(), readMixerValues(), 
     toggleMain(), propertyNotify();

/* Callback Prototypes */
void slideCallback(), slideRBCallback(), dialCallback();

/* registered actions. These can be called using translations */

static XtActionsRec actions[] =
{
    {"Messages", messages},
    {"PropertyNotify", propertyNotify},
    {"Quit", quitApplication},
    {"ReadValues", readMixerValues},
    {"EnterWindow", enterWindow},
    {"ToggleMain", toggleMain},
};

/* translations, i.e. the procedures called when specific events occur */

/* translation for the mixer's top level window, i.e. the window that
   is managed by the window manager */
static char top_level_translations[] =
    "<Message>: Messages() \n\
        <Prop>: PropertyNotify() \n\
       <Enter>: EnterWindow() \n\
      <Expose>: ReadValues()";

/* translation for the form_widget. This is a child of the top_level_widget.
   Note the the top_level_widget does not receive keyboard events */
static char form_translations[] =
    "<Key>q:    Quit()";

/* the top level translations for the launcher window */
static char launch_translations[] =
    "<Message>: Messages() \n\
       <Enter>: EnterWindow() \n\
      <Expose>: ReadValues()";

/* and for its child widget, as the top level widget gets no
   button events */
static char launch_vol_translations[] =
     "<Btn3Up>:  ToggleMain()";

/* fallback resources. These are read when there are no application-defaults */
static String fallback_resources[] =
{
    "*Dial.width:40",
    "*Dial.height:50",
    "*Dial.borderWidth:0",
    "*Dial.dialColor:gray75",
    "*Dial.dialBackColor:white",
    "*Dial.dialBorderColor:black",
    "*Dial.dialPointColor:red",
    "*Dial.dialPointBorderColor:red",
    "*Slide.width:16",
    "*Slide.height:109",
    "*Slide.borderWidth:0",
    "*Slide.slideBackground:black",
    "*Slide.slideColor:red",
    "*Slide.rBBackColor:black",
    "*Slide.rBColor:green",
    "*vol.label:Vol",
    "*balance.label:Bal",
    "*bass.label:Bas",
    "*treble.label:Tre",
    NULL, /* Must be NULL terminated */
};

/* SOUND_DEVICE_NAMES comes from <linux/soundcard.h> */
char *sound_device_names[] = SOUND_DEVICE_NAMES; /* widget names */

/* program name, from argv[0] */
char *program_name;

/* mixer flags */
int supported, stereo_devs, rec_mask, error_field;

/* mixer device file descriptor */
int mixer_fd;

/* no. of present mixer channels */
int nr_of_devices;

/* mapping state of top level widget */
int top_level_mapped; 

/* every mixer channel gets a widget */
Widget mixer_widgets[SOUND_MIXER_NRDEVICES];

/* additional widgets */
Widget application_widget, top_level_widget, form_widget, launch_widget, 
    launch_vol_widget;

/* interclient communications stuff */
static Atom wm_protocols, delete_window, save_yourself, wm_state;

/* KDE flag */
Boolean kde_running;

/* mixer symbols for restore values */
static char *charmap="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdef";

/* forward declaration for session management */
void readStringFromMixer();

/*
 * Utility function. Returns the geometry of widget w as seen from the
 * root window. Used for session management, so we can supply a
 * -geometry argument to place the mixer window exactly where we left it.
 */
char *getGeo(Widget w, char *result) {
    int x, y, x1, y1;
    unsigned int width, height, border_width, depth,
        width1, height1, border_width1, depth1, childno;
    Window current, root, parent, *children;

    current = XtWindow(w);
    XGetGeometry(XtDisplay(w), current, &root, &x, &y, 
                 &width, &height, &border_width, &depth);
    /* we must descend the window hierarchy until we get to the root window */
    do {
        XQueryTree(XtDisplay(w), current, &root, &parent, &children, &childno);
        if (childno>0) XFree(children);
        current=parent;
        if (current!=root) {
            XGetGeometry(XtDisplay(w), current, &root, &x1, &y1, 
                 &width1, &height1, &border_width1, &depth1);
            x=x1;
            y=y1;
        }
    } while (current != root);
    sprintf(result, "%dx%d+%d+%d", width, height, x, y);
    return result;
}

/*
 * The launch window can toggle the mixer window on and off using this
 * function.
 */
/* ARGSUSED */
void toggleMain(Widget w, XButtonEvent *event,
          String *params, Cardinal *count) {
    Atom a;
    long data;

    if (!top_level_mapped) { /* map window */
        /* KDE needs extra care */
        if ((kde_running) &&
            ((a = XInternAtom(XtDisplay(top_level_widget),
                              "KWM_WIN_ICONIFIED", False)) != None)) {
            data = 0; /* not iconified */
            XChangeProperty(XtDisplay(top_level_widget), 
                            XtWindow(top_level_widget), a, a, 32,
                            PropModeReplace, (unsigned char *)&data, 1);
        }        
        top_level_mapped=1;
        XtMapWidget(top_level_widget);
    }
    else { /* unmap window */
        top_level_mapped=0;
        XtUnmapWidget(top_level_widget);
    }
}

/*
 * Collect messages from other applications, especially the window
 * manager. This includes the DELETE_WINDOW and the SAVE_YOURSELF
 * protocols. The KDEchangePalette cannot be handled here because
 * the event is sent to a window Xt knows nothing about.
 */
/* ARGSUSED */
void messages(Widget w, XClientMessageEvent *event,
              String *params, Cardinal *count) {
    char real_path_name[256], geometry[256], *cmds[6];
    char restore_value[256];
    int i;
    Boolean iconic;

    if (event->message_type == wm_protocols) {
        /* delete window */
        if ((Atom) event->data.l[0] == delete_window) {
            if ((app_data.launcher) && 
                (XtWindow(w) == XtWindow(top_level_widget))) { 
                toggleMain(w, NULL, params, count);
            }
            else exit(0);
        }      
        /* save yourself */
        if ((Atom) event->data.l[0] == save_yourself) {
            if (realpath(program_name, real_path_name) == NULL) {
                strcpy(real_path_name, program_name);
            }
            getGeo(top_level_widget, geometry);
            readStringFromMixer(restore_value);
            XtVaGetValues(top_level_widget, XtNiconic, &iconic, NULL);
            i=0;
            cmds[i]=real_path_name; i++;
            cmds[i]="-geometry"; i++;
            cmds[i]=geometry; i++;
            cmds[i]="-restore"; i++;
            cmds[i]=restore_value; i++;
            if (iconic) {
                cmds[i]="-iconic"; i++;
            }
            XSetCommand(XtDisplay(w), XtWindow(w), cmds, i);
        }
    }
}

/*
 * quit this application or close mixer window 
 */
/* ARGSUSED */
void quitApplication(Widget w, XKeyEvent *event,
          String *params, Cardinal *count) {

    if (app_data.launcher) { 
        toggleMain(w, NULL, params, count);
    }
    else {
        exit(0);
    }
}

/* 
 * When a launcher window exists, there is no need to iconify the
 * mixer window. But we cannot tell the window manager that it should
 * omit the iconify button. So we intercept this procedure and close
 * (in this case: withdraw) the window when it gets iconified.
 */
/* ARGSUSED */
void propertyNotify(Widget w, XPropertyEvent *event,
          String *params, Cardinal *count) {
    unsigned char *value;
    Atom a;
    int format;
    unsigned long items_returned, bytes_returned;

    if ((app_data.launcher) 
        && (XtWindow(w)==XtWindow(top_level_widget))
        && (event->atom==wm_state)) {
        /* find out about the window's state */
        XGetWindowProperty(XtDisplay(w), XtWindow(w), wm_state, 0, 1,
                           False, AnyPropertyType, &a, &format,
                           &items_returned, &bytes_returned, &value);
        if ((a!=None) && (bytes_returned>=sizeof(long)) &&
        ((long) *value == IconicState) && (top_level_mapped)) { 
            top_level_mapped=0;
            XWithdrawWindow(XtDisplay(w), XtWindow(w), 
                            XScreenNumberOfScreen(XtScreen(w)));
        }
        if (bytes_returned>0) XFree(value);
    }
}

/* 
 * Having more than one program accessing (and changing) the sound card's
 * mixer values, the shown values are not always up to date. We could
 * poll the mixer every second or so, but I think it's sufficient to
 * update the values when the mouse enters the application window.
 */
/* ARGSUSED */
void enterWindow(Widget w, XCrossingEvent *event,
                 String *params, Cardinal *count) {

    readMixerValues();
}

/* print a fatal error message and exit*/
void fatal() {

    fprintf(stderr, "FATAL: your sound board probably does not have a mixer device\n");
    exit(-1);
}

/* decribe the command line options and exit */
void info() {
    
    fprintf(stderr, "USAGE: %s [ X toolkit options ...] [ options ...]\n"
            ,program_name);
    fprintf(stderr, "xtmix %s - an X Window interface to the Linux Sound Driver Mixer\n",VERSION);
    fprintf(stderr, "options include:\n");
    fprintf(stderr, "    -dev <device>            specify mixer device other than /dev/mixer\n");
    fprintf(stderr, "    -nofilter                omit treble / bass dials\n");
    
    fprintf(stderr, "    -launch                  start up a small launcher window that can be\n");
    fprintf(stderr, "                             'swallowed' by a toolbar\n");
    fprintf(stderr, "    -lbg                     specify the launcher window's background\n");
    fprintf(stderr, "    -llabel                  specify the launcher window's label\n");    
    fprintf(stderr, "    -ignoreKDE               ignore the existence of KDE\n");
    fprintf(stderr, "(c) 1995, 1997, 1998 by Martin Denn <Martin.Denn@rhein-main.de>\n");
#if defined(COMPILETIME) && defined(COMPILENAME)
    fprintf(stderr, "compiled %s by %s\n",COMPILETIME,COMPILENAME);
#endif
    exit(-1);
}

/*
 * open the mixer device and find out about supported channels,
 * stereo devices, recording mask etc.
 */
void mixerInit() {

    mixer_fd = open (app_data.mixer_device, O_RDWR, 0);
    if (mixer_fd < 0 ) {
        fprintf (stderr, "%s: Error opening mixer device %s\n",
                 program_name, app_data.mixer_device);
        fprintf (stderr, "FATAL: %s probably does not exist or has no r/w permissions\n", app_data.mixer_device);
        exit (-1);
    }
    if (ioctl(mixer_fd, SOUND_MIXER_READ_DEVMASK, &supported) == -1)
        supported = ((1 << (SOUND_MIXER_NRDEVICES))-1);

    if (ioctl(mixer_fd, SOUND_MIXER_READ_STEREODEVS, &stereo_devs) == -1)
        stereo_devs = ((1 << (SOUND_MIXER_NRDEVICES))-1);

    if (ioctl(mixer_fd, SOUND_MIXER_READ_RECMASK, &rec_mask) == -1)
        rec_mask = ((1 << (SOUND_MIXER_NRDEVICES))-1);

    if (app_data.no_filters) {
        supported = supported & ~(1 << SOUND_MIXER_BASS);
        supported = supported & ~(1 << SOUND_MIXER_TREBLE);
    }   
    /* error field contains all supported channels. Whenever we get 
       an error accessing a certain channel, this channel gets disabled.
       When all channels are disabled, the application exits */
    error_field = supported;
}

/* build pixmaps using the Xpm library */
void createPixmap(Display *dpy, int i, Pixmap *pix, Pixmap *mask) {
    char **data;
    XpmAttributes attributes;

    attributes.valuemask = XpmSize;
    switch (i) {
        case MINI_ICON:
            data = xtmix_mini_xpm;
            break;
        case SOUND_MIXER_MIC:
            data = mic_xpm;
            break;
        case SOUND_MIXER_CD:
            data = cd_xpm;
            break;
        case SOUND_MIXER_SYNTH:
            data = synth_xpm;
            break;
        case SOUND_MIXER_LINE:
            data = line_xpm;
            break;
        case SOUND_MIXER_PCM:
            data = pcm_xpm;
            break;
        case SOUND_MIXER_SPEAKER:
            data = speaker_xpm;
            break;
/* Well, the next pictures are rather ugly */
        case SOUND_MIXER_IMIX:
            data = mix_xpm;
            break;
        case SOUND_MIXER_ALTPCM:
            data = pcm2_xpm;
            break;
        case SOUND_MIXER_RECLEV:
            data = rec_xpm;
            break;
        case SOUND_MIXER_IGAIN:
            data = igain_xpm;
            break;
        case SOUND_MIXER_OGAIN:
            data = ogain_xpm;
            break;
        case SOUND_MIXER_LINE1:
            data = line1_xpm;
            break;
        case SOUND_MIXER_LINE2:
            data = line2_xpm;
            break;
        case SOUND_MIXER_LINE3:
            data = line3_xpm;
            break;
             
/* (there was no need for it, because I only own an SB Pro). So, if someone 
 * wants to create better pictures, feel free to send them to me: 
 * Martin.Denn@rhein-main.de
 * Thanks :-)
 */
        default:
            data = empty_xpm;
    }
    if (data != NULL)
        XpmCreatePixmapFromData(dpy, DefaultRootWindow(dpy),
                                data, pix, mask, &attributes);
    else {
        (*pix) = XtUnspecifiedPixmap;
    }
    
    if (mask == NULL) {
        (*mask) = XtUnspecifiedPixmap;
    }
}

/* build the top level widget and initialize the application */
void createApplication(int *argc, char **argv, XtAppContext *app_context,
                       char *class, Widget *top) {

    (*top) = XtVaAppInitialize(
        app_context,                /* Application context */
	class,                      /* Application class */
        options, XtNumber(options), /* command line option list */
        argc, argv,                 /* command line args */
        fallback_resources,         /* for missing app-defaults file */
        NULL);                      /* terminate varargs list */

    XtVaGetApplicationResources(
        (*top),                         /* widget */
        &app_data,                      /* data field */
        resources, XtNumber(resources), /* resources */
        NULL);                          /* terminate varargs list */
    
    mixerInit();

    /* check for main volume control */
    if ((((1 << SOUND_MIXER_VOLUME) & supported) == 0) 
        && (app_data.launcher)) {
        fprintf(stderr, "%s: This mixer has no main volume control, therefore a launcher window cannot be provided\n", program_name);
        app_data.launcher=False;
    }

    /* any options left? */
    if ((*argc) > 1) info();
}

/* create an independent top level widget */
void createPopup(Widget parent, char *title, Widget *top) {

    (*top) = XtVaCreatePopupShell(title, topLevelShellWidgetClass,
                                parent, NULL);
}

/* create the widgets to control the different mixer channels */
void createMixerWidgets(Widget top) {
    int i, n = 0;
    char *left = NULL;
    Pixmap pix, mask;
    Boolean RB, common;
    
    /* this widget manages all channel control widgets */
    form_widget = XtVaCreateManagedWidget("form", formWidgetClass, top, NULL);

    for (i = 0; i < SOUND_MIXER_NRDEVICES; i++) {
        if ((i != SOUND_MIXER_VOLUME) && (((1 << i) & supported) != 0) &&
            (i != SOUND_MIXER_BASS) && (i != SOUND_MIXER_TREBLE)) {
            /* we create a slide bar: */
            createPixmap(XtDisplay(top),i, &pix, &mask);
            RB = (((1 << i) & rec_mask) != 0);
            common = (((1 << i) & stereo_devs) == 0);
            if (left == NULL) {
                mixer_widgets[n]=XtVaCreateManagedWidget(sound_device_names[i],
                                               slideWidgetClass, form_widget,
                                               XtNpixmap, pix,
                                               XtNclipmask, mask,
                                               XtNrBExists, RB,
                                               XtNcommon, common,
                                               NULL);
            }
            else {
                mixer_widgets[n]=XtVaCreateManagedWidget(sound_device_names[i],
                                               slideWidgetClass, form_widget,
                                               XtVaTypedArg, XtNfromHoriz, 
                                               XtRString, left,
                                               strlen(left)+1,
                                               XtNpixmap, pix,
                                               XtNclipmask, mask,
                                               XtNrBExists, RB,
                                               XtNcommon, common,
                                               NULL);
            }
            left = sound_device_names[i];
            XtAddCallback(mixer_widgets[n], XtNcallback, slideCallback, 
                          (XtPointer) i);
           if (RB)
                XtAddCallback(mixer_widgets[n], XtNrBCallback, 
                              slideRBCallback, (XtPointer) i);
            n++;
        }
    } 
    /* the rest will get dial buttons */
    i = SOUND_MIXER_VOLUME;
    if (((1 << i) & supported) != 0) {
        mixer_widgets[n] = XtVaCreateManagedWidget(sound_device_names[i],
                                       dialWidgetClass, form_widget,
                                       XtVaTypedArg, XtNfromHoriz, 
                                       XtRString, left,
                                       strlen(left)+1,
                                       NULL);
        XtAddCallback(mixer_widgets[n], XtNcallback, dialCallback, 
                      (XtPointer) i);
        n++;
        mixer_widgets[n] = XtVaCreateManagedWidget("balance",
                                       dialWidgetClass, form_widget,
                                       XtVaTypedArg, XtNfromHoriz, 
                                       XtRString, left,
                                       strlen(left)+1,
                                       XtVaTypedArg, XtNfromVert, 
                                       XtRString, sound_device_names[i],
                                       strlen(sound_device_names[i])+1,
                                       NULL);
        XtAddCallback(mixer_widgets[n], XtNcallback, dialCallback, 
                      (XtPointer)BALANCE);
        left = sound_device_names[i];
        n++;

    }
    i = SOUND_MIXER_BASS;
    if (((1 << i) & supported) != 0) {
        mixer_widgets[n] = XtVaCreateManagedWidget(sound_device_names[i],
                                       dialWidgetClass, form_widget,
                                       XtVaTypedArg, XtNfromHoriz, 
                                       XtRString, left,
                                       strlen(left)+1,
                                       NULL);
        XtAddCallback(mixer_widgets[n], XtNcallback, dialCallback, 
                      (XtPointer) i);
        n++;
    }
    i = SOUND_MIXER_TREBLE;
    if (((1 << i) & supported) != 0) {
        mixer_widgets[n] = XtVaCreateManagedWidget(sound_device_names[i],
                                       dialWidgetClass, form_widget,
                                       XtVaTypedArg, XtNfromHoriz, 
                                       XtRString, left,
                                       strlen(left)+1,
                                       XtVaTypedArg, XtNfromVert, 
                                       XtRString, sound_device_names[SOUND_MIXER_BASS],
                                       strlen(sound_device_names[SOUND_MIXER_BASS])+1,
                                       NULL);
        XtAddCallback(mixer_widgets[n], XtNcallback, dialCallback, 
                      (XtPointer) i);
        n++;
    }
    nr_of_devices = n;    
}

/* create the launchers volume control widget */ 
void createLauncher(Widget top) {
    int i;

    i = SOUND_MIXER_VOLUME;
    if (((1 << i) & supported) != 0) {
        launch_vol_widget = XtVaCreateManagedWidget("launchVolume",
                                       dialWidgetClass, top,
                                       NULL);
        XtAddCallback(launch_vol_widget, XtNcallback, dialCallback, 
                      (XtPointer) i);
    }
}

/* 
 * Whenever a slide is moved, this function will be called.
 */
/* ARGSUSED */
void slideCallback(Widget w, XtPointer client_data, XtPointer call_data) {
    int dev_nr = (int) client_data;
    SlideInfo *info = (SlideInfo *) call_data;
    int value;
    
    value = ( info->lvalue | ((info->rvalue) << 8));
    if (ioctl(mixer_fd,MIXER_WRITE(dev_nr),&value) == -1) {
        fprintf(stderr,"%s: Error writing %s in %s\n",
                       program_name, sound_device_names[dev_nr],
                       app_data.mixer_device);
        error_field = (error_field & ~(1 << dev_nr));
        if (error_field == 0) fatal();
    }
}

/*
 * This function is called when the recording source is changed.
 */
/* ARGSUSED */
void slideRBCallback(Widget w, XtPointer client_data, XtPointer call_data) {
    int dev_nr = (int) client_data;
    int value, cur_value, i;

    /* get current recording source mask */
    if (ioctl(mixer_fd,SOUND_MIXER_READ_RECSRC,&cur_value) == -1) {
        fprintf(stderr,"%s: Error reading recording source\n",
                      program_name);
    }
    /* set new mask */
    value = (cur_value ^ (1 << dev_nr));
    if (ioctl(mixer_fd,SOUND_MIXER_WRITE_RECSRC,&value) == -1) {
        fprintf(stderr,"%s: Error setting recording source\n",
                      program_name);
    }
    /* reread recording source mask */
    if (ioctl(mixer_fd,SOUND_MIXER_READ_RECSRC,&value) == -1) {
        fprintf(stderr,"%s: Error reading recording source\n",
                      program_name);
    }
    /* show changes */
    for (i=0; i<SOUND_MIXER_NRDEVICES; i++) {
        if ((1 << i) & supported) {
            XtVaSetValues(XtNameToWidget(form_widget, sound_device_names[i]),
                          XtNrBOn, (Boolean) ((value & (1 << i)) != 0),
                          NULL);
        }
    }
}

/*
 * when you move the dial button, this procedure is called. The
 * balance/volume stuff seems rather complicated; I used this function:
 *
 * (Percentage of volume)
 *   100%            xxxxxx|xxxxxxxxxxxxxxxx 
 *               xxxx      |
 *            xxx          |
 *           x             |
 *    50%   x              |                  (right channel, 
 *         x               |                   left is vice versa)
 *         x               |
 *        x                |
 *     0% x                |
 *        +----------------+----------------+
 *      -100               0               100 (Balance Value)
 */
/* ARGSUSED */
void dialCallback(Widget w, XtPointer client_data, XtPointer call_data) {
    int dev_nr = (int) client_data;
    DialInfo *info = (DialInfo *) call_data;
    Widget update;
    int value, n, vol, bal, left_vol, right_vol;
    
    n=0;
    switch (dev_nr) {
        case SOUND_MIXER_BASS:
        case SOUND_MIXER_TREBLE:
            value = ( info->value | ((info->value) << 8));
            n=dev_nr;
            break;
        case SOUND_MIXER_VOLUME:
        case BALANCE:
            if (dev_nr==SOUND_MIXER_VOLUME) {
                vol=info->value;
                XtVaGetValues(XtNameToWidget(form_widget, "balance"),
                              XtNcurValue, &bal, NULL);
            }
            if (dev_nr==BALANCE) {
                bal=info->value;
                XtVaGetValues(XtNameToWidget(form_widget,
                              sound_device_names[SOUND_MIXER_VOLUME]),
                              XtNcurValue, &vol, NULL);
            }
            left_vol = (int) ((bal<50)?50:CIRCLE((bal)))*vol/50.0;
            right_vol= (int) ((bal>50)?50:CIRCLE((bal)))*vol/50.0;
            value = ( left_vol | ((right_vol) << 8));
            n=SOUND_MIXER_VOLUME;
            break; 
        default: 
            break;
    }
    if (ioctl(mixer_fd, MIXER_WRITE(n), &value) == -1) {
        fprintf(stderr,"%s: Error writing %s in %s\n",
                       program_name, sound_device_names[n],
                       app_data.mixer_device);
        error_field = (error_field & ~(1 << n));
        if (error_field == 0) fatal();
    }
    /* don't forget to update the launcher widget */
    if ((app_data.launcher) && (dev_nr==SOUND_MIXER_VOLUME)) {
        if (XtWindow(w)==XtWindow(launch_vol_widget))
            update=XtNameToWidget(form_widget,
                                  sound_device_names[SOUND_MIXER_VOLUME]);
        else
            update=launch_vol_widget;
        XtVaSetValues(update, XtNcurValue, vol, NULL);
    }
}

/* read mixer and put all relevant values into a string representation */
void readStringFromMixer(char *result) {
    int i, value;

    if (ioctl(mixer_fd, SOUND_MIXER_READ_RECSRC, &value) == -1) {
        fprintf(stderr,"%s: Error reading recording source\n", program_name);
    }
    sprintf(result, "%c%04x", charmap[31], value);
    for (i=0; i<SOUND_MIXER_NRDEVICES; i++) {
        if (((1 << i) & supported) != 0) {
            if (ioctl(mixer_fd, MIXER_READ(i), &value) == -1) {
                fprintf(stderr,"%s: Error reading %s in %s\n",
                        program_name, sound_device_names[i],
                        app_data.mixer_device);
                error_field = (error_field & ~(1 << i));
                if (error_field == 0) fatal();
            }
            else {
                sprintf(result, "%s%c%04x", result, charmap[i], value);
            }
        }
    }
    printf("%s\n", result);
}
        
/* set the mixer according to a given string */
void writeStringToMixer(char *values) {
    int i, v;
    char *c, value[4];

    if (strlen(values)%5 != 0) {
        fprintf(stderr,"%s: -restore argument invalid\n", program_name);
    }
    c=values;
    while ((*c)!='\0') {
        i=0; while ((i<32) && (charmap[i]!=(*c))) i++;
        if (i>=32) {
            fprintf(stderr,"%s: -restore argument invalid\n", program_name);
        }
        c++;
        strncpy(value, c, 4); sscanf(value, "%x", &v);
        c+=4;
        if (i==31) {
            if (ioctl(mixer_fd, SOUND_MIXER_WRITE_RECSRC, &v) == -1) {
                fprintf(stderr,"%s: Error writing recording source\n", program_name);
            }
        }
        else if (((1 << i) & supported) != 0) {
            if (ioctl(mixer_fd, MIXER_WRITE(i), &v) == -1) {
                fprintf(stderr,"%s: Error writing %s in %s\n",
                        program_name, sound_device_names[i],
                        app_data.mixer_device);
                error_field = (error_field & ~(1 << i));
                if (error_field == 0) fatal();
            }
        }
    }
}

/* this procedure reads the values from the mixer device */
void readMixerValues() {
    int i, value, lvalue, rvalue, vol, bal, left_vol, right_vol, rec_src;


    if (ioctl(mixer_fd, SOUND_MIXER_READ_RECSRC, &rec_src) == -1) {
        fprintf(stderr,"%s: Error reading recording source\n", program_name);
    }
        
    for (i=0; i<SOUND_MIXER_NRDEVICES; i++) {
        if (((1 << i) & supported) != 0) {
            if (ioctl(mixer_fd, MIXER_READ(i), &value) == -1) {
                fprintf(stderr,"%s: Error reading %s in %s\n",
                        program_name, sound_device_names[i],
                        app_data.mixer_device);
                error_field = (error_field & ~(1 << i));
                if (error_field == 0) fatal();
            }
            else {
                lvalue=(value & 0x00FF);
                rvalue=((value & 0xFF00) >> 8);
                
                switch (i) {
                    case SOUND_MIXER_VOLUME:
                        vol = MAX((lvalue),(rvalue));
                        XtVaSetValues(XtNameToWidget(form_widget, 
                                      sound_device_names[i]),
                                      XtNcurValue, vol, NULL);
                        if (app_data.launcher) {
                            XtVaSetValues(launch_vol_widget,
                                          XtNcurValue, vol, NULL);
                        }
                        
                        /* test old values */
                        XtVaGetValues(XtNameToWidget(form_widget, "balance"),
                                      XtNcurValue, &bal, NULL);

                        left_vol = (int) ((bal<50)?50:CIRCLE((bal)))*vol/50.0;
                        right_vol= (int) ((bal>50)?50:CIRCLE((bal)))*vol/50.0;
                        /* if old values are incorrect -> calculate new ones */
                        if ((abs(left_vol-lvalue)>CORRECTION)
                           ||(abs(right_vol-rvalue)>CORRECTION)) {
                            if ((vol==0)||
                                ((vol>0)&&(abs(lvalue-rvalue)<CORRECTION))) 
                                bal=50;
                            else if ((vol>0)&&(lvalue<rvalue)) {
                                bal=50+CIRCLE((50-lvalue*50/rvalue));
                            }
                            else {
                                bal=50-CIRCLE((50-rvalue*50/lvalue));
                            }
                            XtVaSetValues(XtNameToWidget(form_widget, 
                                                         "balance"),
                                              XtNcurValue, bal, NULL);
                        }
                        break;
                    case SOUND_MIXER_BASS:
                    case SOUND_MIXER_TREBLE:
                        XtVaSetValues(XtNameToWidget(form_widget,
                                      sound_device_names[i]),
                                      XtNcurValue, (lvalue+rvalue)/2,
                                      NULL);
                        break;
                    default:
                        XtVaSetValues(XtNameToWidget(form_widget, 
                                      sound_device_names[i]),
                                      XtNleftValue, lvalue,
                                      XtNrightValue, rvalue,
                                      XtNrBOn, (((1 << i) & rec_src) != 0),
                                      NULL);
                        break;
                }
            }
        }
    }           
}

void realizeWidget(Widget w) {
    char *title;
    Boolean iconic;
    int icon_x, icon_y, initial_state;
    Window root, dummy;
    int x, y, parse_mask;
    int user_x, user_y;
    unsigned int width, height, border_width, depth;
    unsigned int user_width, user_height;
    unsigned int root_width, root_height;
    char *geometry;
    unsigned int gravity;
    XSizeHints *size_hints;
   
    /* get resources from application_widget (which will never be mapped).
       Here we get the value from the -geometry option */
    XtVaGetValues(application_widget, XtNiconX, &icon_x, XtNiconY, &icon_y,
                  XtNiconic, &iconic, XtNinitialState, &initial_state,
                  XtNtitle, &title, XtNgeometry, &geometry, NULL);
 
    /* realize widget */
    XtRealizeWidget(w);
    /* get geometry. This is the minimal geometry, the normal size
       when no -geometry option is set */
    XGetGeometry(XtDisplay(w), XtWindow(w),
                 &root, &x, &y, &width, &height, &border_width, &depth);

    /* set non-geometry resources */
    XtVaSetValues(w, XtNiconX, icon_x, XtNiconY, icon_y,
                  XtNiconic, iconic, XtNinitialState, initial_state,
                  XtNtitle, title, NULL);

    /* get the size of the root window */
    XGetGeometry(XtDisplay(w), root,
                 &dummy, &x, &y, &root_width, &root_height, 
                 &border_width, &depth);
    
    /* initialize WindowManager hints */
    gravity = 0;
    size_hints = XAllocSizeHints();

    /* set minimal geometry */
    size_hints->min_width = width;
    size_hints->min_height = height;
    size_hints->flags = PMinSize;

    /* if a -geometry option is specified: parse the arguments and
       set the values */
    if (geometry != NULL) {
        parse_mask = XParseGeometry(geometry, &user_x, &user_y, 
                                    &user_width, &user_height);
 
        if (parse_mask & WidthValue) {
            if (user_width < size_hints->min_width) {
                fprintf(stderr, "user specified width too small, setting to minimal allowed value\n");
                width = size_hints->min_width;
            }
            else
                width = user_width;
            XtVaSetValues(w, XtNwidth, width, NULL);
            size_hints->width = width;
            size_hints->flags |= USSize;
        }
        if (parse_mask & HeightValue) {
            if (user_height < size_hints->min_height) {
                fprintf(stderr, "user specified height too small, setting to minimal allowed value\n");
                height = size_hints->min_height;
            }
            else
                height = user_height;
            XtVaSetValues(w, XtNheight, height, NULL);
            size_hints->height = height;
            size_hints->flags |= USSize;
        }
        if (parse_mask & XValue) {
            if (parse_mask & XNegative) {
                user_x+=root_width-width;
                gravity |= 0x01;
            }
            XtVaSetValues(w, XtNx, user_x, NULL);
            size_hints->x = user_x;
            size_hints->flags |= USPosition;
        }
        if (parse_mask & YValue) {
            if (parse_mask & YNegative) {
                user_y+=root_height-height;
                gravity |= 0x02;
            }
            XtVaSetValues(w, XtNy, user_y, NULL);
            size_hints->y = user_y;
            size_hints->flags |= USPosition;
        }
        switch (gravity) {
        case 0x00: size_hints->win_gravity = NorthWestGravity;
            break;
        case 0x01: size_hints->win_gravity = NorthEastGravity;
            break;
        case 0x02: size_hints->win_gravity = SouthWestGravity;
            break;
        case 0x03: size_hints->win_gravity = SouthEastGravity;
            break;
        }
        size_hints->flags |= PWinGravity;
    }
    /* set size hints */
    XSetWMNormalHints(XtDisplay(w), XtWindow(w), size_hints);
    XFree(size_hints);
}
    
    

/* put values from -lbg and -llabel options into resource database */
void setLauncherValues(Widget w) {
    XrmDatabase db, app_db;
    char *db_line;
    char *label="xtmix.xtmix-launcher.launchVolume.label";
    char *background="xtmix.xtmix-launcher.launchVolume.background";

    db=NULL;
    if (app_data.launcher_label != NULL) {
        db_line = (char *) 
            malloc(strlen(label)+strlen(app_data.launcher_label)+2);
        sprintf(db_line, "%s:%s", label, app_data.launcher_label);
        XrmPutLineResource(&db, db_line);
    }
    if (app_data.launcher_background != NULL) {
        db_line = (char *) 
            malloc(strlen(background)+strlen(app_data.launcher_background)+2);
        sprintf(db_line, "%s:%s", background, app_data.launcher_background);
        XrmPutLineResource(&db, db_line);
    }
    app_db=XtDatabase(XtDisplay(w));
    XrmMergeDatabases(db, &app_db);
}

/* main routine */
void main(int argc, char **argv) {
    XtAppContext app_context;
    XtTranslations trans;
    Pixmap icon_pixmap, mini_icon, mini_icon_mask;
    Atom wm_protos[2];
    Window root;
    Display *dpy;
    XEvent ev;
    char *application_class = "XtMix";
    int i;
    
    program_name = argv[0];

    /* init mixer and widgets */
    createApplication(&argc, argv, &app_context, application_class,
                      &application_widget);
    createPopup(application_widget, "xtmix", &top_level_widget);
    createMixerWidgets(top_level_widget);
    if (app_data.launcher) {
        setLauncherValues(application_widget);
        createPopup(application_widget, "xtmix-launcher", &launch_widget);
        createLauncher(launch_widget);
    }

    /* add actions and translations */
    XtAppAddActions(app_context, actions, XtNumber(actions));

    trans = XtParseTranslationTable(top_level_translations);
    XtOverrideTranslations(top_level_widget, trans);
    trans = XtParseTranslationTable(form_translations);
    XtOverrideTranslations(form_widget, trans);
    if (app_data.launcher) {
        trans = XtParseTranslationTable(launch_translations);
        XtOverrideTranslations(launch_widget, trans);
        trans = XtParseTranslationTable(launch_vol_translations);
        XtOverrideTranslations(launch_vol_widget, trans);
    }
 
    dpy = XtDisplay(top_level_widget);
    root = RootWindowOfScreen(XtScreen(top_level_widget)),

    /* add icon pixmap, this is the old way to supply a monochrome
       bitmap */
    icon_pixmap = XCreateBitmapFromData(dpy, root,
			 xtmix_bits, xtmix_width, xtmix_height);
    XtVaSetValues(top_level_widget, XtNiconPixmap, icon_pixmap, NULL);

    if (app_data.restore != NULL) {
        writeStringToMixer(app_data.restore);
    }

    /* read initial mixer values */
    readMixerValues();
    
    /* realize and map Widgets */
    realizeWidget(top_level_widget);
    if (app_data.launcher) {
        XtRealizeWidget(launch_widget);
        XtMapWidget(launch_widget);
    }
    else {
        XtMapWidget(top_level_widget);
    }
    
    
    /* add the "delete" and "save_yourself" - features to xtmix */
    wm_protocols = XInternAtom(dpy, "WM_PROTOCOLS", False);
    wm_state = XInternAtom(dpy, "WM_STATE", False);
    i=0;
    delete_window = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
    if (delete_window != None) wm_protos[i++] = delete_window;
    if (!app_data.launcher) {
        /* When the launcher is active, we need no session management,
           because the window manager who 'swallows' the launcher
           window will start the mixer automatically */
        save_yourself = XInternAtom(dpy, "WM_SAVE_YOURSELF", False);
        if (save_yourself != None) wm_protos[i++] = save_yourself;
    }
    XSetWMProtocols(dpy, XtWindow(top_level_widget), wm_protos, i);

    /* get KDE aware */
    kde_running = (detectKDE(application_widget) && (!app_data.ignore_KDE));
    if (kde_running) {
#ifdef KDE_BETA_3
        registerResourceTranslation(application_widget,
                                    "*background: BackgroundColor\n\
                                     *labelColor: TextColor");
#else
        registerResourceTranslation(application_widget,
                                    "*background: background\n\
                                     *labelColor: foreground");
#endif       
        createPixmap(dpy, MINI_ICON, &mini_icon, &mini_icon_mask);
        registerMiniIcon(top_level_widget, mini_icon, mini_icon_mask);
    }
    /* The main loop */
    while (1) {
        XtAppNextEvent(app_context, &ev);
        /* special event processing */
        if (kde_running)
            processKDEEvents(ev);
        XtDispatchEvent(&ev);
    }
}



















